# ICASSP experient - Performances

include("../../src/BnbScreening.jl")
using StatsBase

################################ Configuration #################################

# This is the only part of code you may have to edit

setup = "toeplitz"          # Setup to use : "gaussian" or "toeplitz"
k = 7                       # Sparsity level
m, n = 500, 300            # Dictionnary dimension
snr = 10. * log10(10.)      # SNR of the observation noise
maxtime = 1000              # Maximum solution time (in second) allowed
maxiter = 1000              # Maximum Active-Set iterations allowed per relaxation resolution
tol = 1.e-8                 # Integer tolerance (ie, |x|<tol <=> x=0)
repeats = 100               # Number of repeats in the experiment


################################## Experiment ##################################

function precompilation(
    setup::String,
    k::Int,
    m::Int,
    n::Int,
    snr::Float64,
    maxtime::Int,
    maxiter::Int,
    tol::Float64,
    )
    println("Precomiling methods ...")
    try
        if setup == "gaussian"
            A, y, λ, M = data_gaussian(k,m,n,snr)
        elseif setup == "toeplitz"
            A, y, λ, M = data_toeplitz(k,m,n,snr)
        else
            error("Parameter `setup` must be either \"gaussian\" or \"toeplitz\".")
        end
        result_mip = solve_mip(A,y,λ,M,mipparams=MipParams(maxtime=maxtime))
        result_bnb = solve_bnb(A,y,λ,M,bnbparams=BnbParams(maxtime=maxtime,maxiter=maxiter,tol=tol,verbosity=false))
        result_bnbs = solve_bnb(A,y,λ,M,bnbparams=BnbParams(maxtime=maxtime,maxiter=maxiter,tol=tol,verbosity=false,screening=true))
    catch e
        e == InterruptException() && rethrow(e)
        println("Precompilation failed with the following error")
        throw(e)
    end
    return nothing
end

logfile_name = string(
    @__DIR__,
    "/",
    Dates.format(Dates.now(),"yyyy-mm-dd-HH:MM:SS"),
    ".txt"
    )
logfile = open(logfile_name, "w")

println("--------------------------------")
println("ICASSP experiment - Performances\n")
println("Configuration")
println("  setup   : $(setup)")
println("  k       : $(k)")
println("  m, n    : $(m), $(n)")
println("  snr     : $(snr)")
println("  maxtime : $(maxtime)")
println("  maxiter : $(maxiter)")
println("  tol     : $(tol)")
println("  repeats : $(repeats)")
println()

write(logfile,"ICASSP experiment - Performances\n\n")
write(logfile,"Configuration\n")
write(logfile,"  setup   : $(setup)\n")
write(logfile,"  k       : $(k)\n")
write(logfile,"  m, n    : $(m), $(n)\n")
write(logfile,"  snr     : $(snr)\n")
write(logfile,"  maxtime : $(maxtime)\n")
write(logfile,"  maxiter : $(maxiter)\n")
write(logfile,"  tol     : $(tol)\n")
write(logfile,"  repeats : $(repeats)\n")

fail_mip = Vector{Union{Bool,Missing}}(missing,repeats)
fail_bnb = Vector{Union{Bool,Missing}}(missing,repeats)
fail_bnbs = Vector{Union{Bool,Missing}}(missing,repeats)
node_mip = Vector{Union{Int64,Missing}}(missing,repeats)
node_bnb = Vector{Union{Int64,Missing}}(missing,repeats)
node_bnbs = Vector{Union{Int64,Missing}}(missing,repeats)
time_mip = Vector{Union{Float64,Missing}}(missing,repeats)
time_bnb = Vector{Union{Float64,Missing}}(missing,repeats)
time_bnbs = Vector{Union{Float64,Missing}}(missing,repeats)

precompilation(setup,k,m,n,snr,maxtime,maxiter,tol)

for i in 1:repeats
    println("Repeat $i/$repeats")
    try
        # Generate data
        if setup == "gaussian"
            A, y, λ, M = data_gaussian(k,m,n,snr)
        elseif setup == "toeplitz"
            A, y, λ, M = data_toeplitz(k,m,n,snr)
        end

        # Run the tree methods
        result_mip = solve_mip(A,y,λ,M,mipparams=MipParams(maxtime=maxtime))
        result_bnb = solve_bnb(A,y,λ,M,bnbparams=BnbParams(maxtime=maxtime,maxiter=maxiter,tol=tol,verbosity=false))
        result_bnbs = solve_bnb(A,y,λ,M,bnbparams=BnbParams(maxtime=maxtime,maxiter=maxiter,tol=tol,verbosity=false,screening=true))

        # Gather results
        if result_mip.termination_status == MOI.OPTIMAL
            fail_mip[i] = false
            node_mip[i] = result_mip.node_count
            time_mip[i] = result_mip.solve_time
        else
            fail_mip[i] = true
        end
        if result_bnb.termination_status == MOI.OPTIMAL
            fail_bnb[i] = false
            node_bnb[i] = result_bnb.node_count
            time_bnb[i] = result_bnb.solve_time
        else
            fail_bnb[i] = true
        end
        if result_bnbs.termination_status == MOI.OPTIMAL
            fail_bnbs[i] = false
            node_bnbs[i] = result_bnbs.node_count
            time_bnbs[i] = result_bnbs.solve_time
        else
            fail_bnbs[i] = true
        end
    catch e
        e == InterruptException() && rethrow(e)
        println("warning : Repeat $i failed with the following error")
        println(e)
    end
end


################################### Results ####################################

println()
println("Fails")
println("  MIP  : $(100 * mean(skipmissing(fail_mip)))%")
println("  BnB  : $(100 * mean(skipmissing(fail_bnb)))%")
println("  SBnB : $(100 * mean(skipmissing(fail_bnbs)))%")
println("Node count")
println("  MIP  : $(mean(skipmissing(node_mip)))")
println("  BnB  : $(mean(skipmissing(node_bnb)))")
println("  SBnB : $(mean(skipmissing(node_bnbs)))")
println("Solve time")
println("  MIP  : $(round(mean(skipmissing(time_mip)),digits=3)) seconds")
println("  BnB  : $(round(mean(skipmissing(time_bnb)),digits=3)) seconds")
println("  SBnB : $(round(mean(skipmissing(time_bnbs)),digits=3)) seconds")
println("\n--------------------------------")

write(logfile,"\n")
write(logfile,"Fails\n")
write(logfile,"  MIP  : $(100 * mean(skipmissing(fail_mip)))%\n")
write(logfile,"  BnB  : $(100 * mean(skipmissing(fail_bnb)))%\n")
write(logfile,"  SBnB : $(100 * mean(skipmissing(fail_bnbs)))%\n")
write(logfile,"Node count\n")
write(logfile,"  MIP  : $(mean(skipmissing(node_mip)))\n")
write(logfile,"  BnB  : $(mean(skipmissing(node_bnb)))\n")
write(logfile,"  SBnB : $(mean(skipmissing(node_bnbs)))\n")
write(logfile,"Solve time\n")
write(logfile,"  MIP  : $(round(mean(skipmissing(time_mip)),digits=3)) seconds\n")
write(logfile,"  BnB  : $(round(mean(skipmissing(time_bnb)),digits=3)) seconds\n")
write(logfile,"  SBnB : $(round(mean(skipmissing(time_bnbs)),digits=3)) seconds\n")
close(logfile)